import contextlib
import datetime
import glob
import os
import random
import shutil
import subprocess
import sys
import tempfile
import traceback
import tiledb
import uuid

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal, assert_equal


def assert_tail_equal(a, *rest, **kwargs):
    """Assert that all arrays in target equal first array"""
    for target in rest:
        assert_array_equal(a, target, **kwargs)


def create_vfs_dir(path):
    """Create a directory at the given scheme-prefixed path,
    first creating the base bucket if needed."""

    import urllib

    split_uri = urllib.parse.urlsplit(path)
    scheme = split_uri.scheme
    bucket = split_uri.netloc
    bucket_uri = scheme + "://" + bucket

    vfs = tiledb.VFS()
    if not vfs.is_bucket(bucket_uri):
        vfs.create_bucket(bucket_uri)
    vfs.create_dir(path)


class DiskTestCase:
    """Helper class to store paths and associated allocation frames. This is both
    a cleanup step and a test of resource management. Some platforms will
    refuse to delete an open file, indicating a potentially leaked resource.
    """

    @classmethod
    def setup_method(self):
        # .lower: because bucket name must be all lowercase
        prefix = "tiledb-" + self.__name__.lower()
        if hasattr(pytest, "tiledb_vfs") and pytest.tiledb_vfs == "s3":
            self.path_scheme = pytest.tiledb_vfs + "://"
            self.rootdir = self.path_scheme + prefix + str(random.randint(0, 10e10))
            create_vfs_dir(self.rootdir)
        else:
            self.path_scheme = ""
            self.rootdir = tempfile.mkdtemp(prefix=prefix)

        self.vfs = tiledb.VFS()
        self.pathmap = dict()

    @classmethod
    def teardown_method(self):
        # Remove every directory starting with rootdir
        # This is both a clean-up step and an implicit test
        # of proper resource deallocation (see notes below)
        for dirpath in glob.glob(self.rootdir + "*"):
            try:
                shutil.rmtree(dirpath)
            except OSError as exc:
                print(
                    "test '{}' error deleting '{}'".format(
                        self.__class__.__name__, dirpath
                    )
                )
                print("registered paths and originating functions:")
                for path, frame in self.pathmap.items():
                    print("  '{}' <- '{}'".format(path, frame))
                raise exc

    def path(self, basename=None, shared=False):
        if self.path_scheme:
            basename = basename if basename else str(uuid.uuid4())
            out = os.path.join(self.rootdir, basename)
            self.vfs.create_dir(out)
        else:
            if basename is not None:
                # Note: this must be `is not None` because we need to match empty string
                out = os.path.abspath(os.path.join(self.rootdir, basename))
            else:
                out = tempfile.mkdtemp(dir=self.rootdir)

        if os.name == "nt" and shared:
            subprocess.run(
                f'cmd //c "net share tiledb-shared={out}"', shell=True, check=True
            )

        # We have had issues in both py and libtiledb in the past
        # where files were not released (eg: destructor not called)
        # Often this is invisible on POSIX platforms, but will
        # cause errors on Windows because two processes cannot access
        # the same file at once.
        # In order to debug this issue, we save the caller where
        # this path was allocated so that we can determine what
        # test created an unreleased file
        frame = traceback.extract_stack(limit=2)[-2][2]
        self.pathmap[out] = frame

        return out

    def assertRaises(self, *args):
        return pytest.raises(*args)

    def assertRaisesRegex(self, e, m):
        return pytest.raises(e, match=m)

    @contextlib.contextmanager
    def assertEqual(self, *args):
        if not len(args) == 2:
            raise Exception("Unexpected input len > 2 to assertEquals")
        assert args[0] == args[1]

    @contextlib.contextmanager
    def assertNotEqual(self, *args):
        if not len(args) == 2:
            raise Exception("Unexpected input len > 2 to assertEquals")
        assert args[0] != args[1]

    @contextlib.contextmanager
    def assertTrue(self, a, msg=None):
        if msg:
            assert a, msg
        else:
            assert a

    @contextlib.contextmanager
    def assertFalse(self, a):
        assert a == False

    @contextlib.contextmanager
    def assertIsInstance(self, v, t):
        assert isinstance(v, t)

    @contextlib.contextmanager
    def assertSetEqual(self, s1, s2):
        assert all(isinstance(x, set) for x in (s1, s2))
        assert s1 == s2

    @contextlib.contextmanager
    def assertIsNone(self, a1):
        assert a1 is None

    @contextlib.contextmanager
    def assertTupleEqual(self, a1, a2):
        assert a1 == a2

    @contextlib.contextmanager
    def assertAlmostEqual(self, a1, a2):
        assert_almost_equal(a1, a2)


# fixture wrapper to use with pytest: mark.parametrize does not
#   work with DiskTestCase subclasses (unittest.TestCase methods
#   cannot take arguments)
@pytest.fixture(scope="class")
def checked_path():
    dtc = DiskTestCase()

    dtc.setup_method()

    yield dtc

    dtc.teardown_method()


# exclude whitespace: if we generate unquoted newline then pandas will be confused
_ws_set = set("\n\t\r")


def gen_chr(max, printable=False):
    while True:
        # TODO we exclude 0x0 here because the key API does not embedded NULL
        s = chr(random.randrange(1, max))
        if printable and (not s.isprintable()) or (s in _ws_set):
            continue
        if len(s) > 0:
            break
    return s


def rand_utf8(size=5):
    return "".join([gen_chr(0xD7FF) for _ in range(0, size)])


def rand_ascii(size=5, printable=False):
    return "".join([gen_chr(127, printable) for _ in range(0, size)])


def rand_ascii_bytes(size=5, printable=False):
    return b"".join([gen_chr(127, printable).encode("utf-8") for _ in range(0, size)])


def dtype_max(dtype):
    if not np.issubdtype(dtype, np.generic):
        raise TypeError("expected numpy dtype!")

    if np.issubdtype(dtype, np.floating):
        finfo = np.finfo(dtype)
        return finfo.max

    elif np.issubdtype(dtype, np.integer):
        iinfo = np.iinfo(dtype)
        return int(iinfo.max)

    elif np.issubdtype(dtype, np.datetime64):
        return np.datetime64(datetime.datetime.max)

    raise "Unknown dtype for dtype_max '{}'".format(str(dtype))


def dtype_min(dtype):
    if not np.issubdtype(dtype, np.generic):
        raise TypeError("expected numpy dtype!")

    if np.issubdtype(dtype, np.floating):
        finfo = np.finfo(dtype)
        return finfo.min

    elif np.issubdtype(dtype, np.integer):
        iinfo = np.iinfo(dtype)
        return int(iinfo.min)

    elif np.issubdtype(dtype, np.datetime64):
        return np.datetime64(datetime.datetime.min)

    raise "Unknown dtype for dtype_min '{dtype}'".format(str(dtype))


def rand_int_sequential(size, dtype=np.uint64):
    arr = np.random.randint(
        dtype_min(dtype), high=dtype_max(dtype), size=size, dtype=dtype
    )
    return np.sort(arr)


def rand_datetime64_array(size, start=None, stop=None, dtype=None):
    if not dtype:
        dtype = np.dtype("M8[ns]")

    # generate randint inbounds on the range of the dtype
    units = np.datetime_data(dtype)[0]
    intmin, intmax = np.iinfo(np.int64).min, np.iinfo(np.int64).max

    if start is None:
        start = np.datetime64(intmin + 1, units)
    else:
        start = np.datetime64(start)
    if stop is None:
        stop = np.datetime64(intmax, units)
    else:
        stop = np.datetime64(stop)

    arr = np.random.randint(
        start.astype(dtype).astype(np.int64),
        stop.astype(dtype).astype(np.int64),
        size=size,
        dtype=np.int64,
    )
    arr.sort()

    return arr.astype(dtype)


def intspace(start, stop, num=50, dtype=np.int64):
    """
    Return evenly spaced values over range ensuring that stop is
    always the maximum (will not overflow with int dtype as linspace)
    :param start:
    :param stop:
    :param num:
    :param dtype:
    :return:
    """
    rval = np.zeros(num, dtype=dtype)
    step = (stop - start) // num
    nextval = start

    if np.issubdtype(dtype, np.integer) and step < 1:
        raise ValueError(
            "Cannot use non-integral step value '{}' for integer dtype!".format(step)
        )

    for i in range(num):
        rval[i] = nextval
        nextval += step

    rval[-1] = stop
    return rval


import pprint as _pprint

pp = _pprint.PrettyPrinter(indent=4)


def xprint(*x):
    for xp in x:
        pp.pprint(xp)


def assert_unordered_equal(a1, a2, unordered=True):
    """Assert that arrays are equal after sorting if
    `unordered==True`"""
    if unordered:
        a1 = np.sort(a1)
        a2 = np.sort(a2)
    assert_array_equal(a1, a2)


def assert_subarrays_equal(a, b, ordered=True):
    assert_equal(a.shape, b.shape)

    if not ordered:
        a = np.sort(a)
        b = np.sort(b)

    for a_el, b_el in zip(a.flat, b.flat):
        assert_array_equal(a_el, b_el)


def assert_all_arrays_equal(*arrays):
    # TODO this should display raise in the calling location if possible
    assert len(arrays) % 2 == 0, "Expected even number of arrays"

    for a1, a2 in zip(arrays[0::2], arrays[1::2]):
        assert_array_equal(a1, a2)


def assert_captured(cap, expected):
    if sys.platform == "win32":
        return
    else:
        import ctypes

        libc = ctypes.CDLL(None)
        libc.fflush(None)

        out, err = cap.readouterr()
        assert not err
        assert expected in out



"""
TODO
- # implement mock of expected behavior in pure numpy w/ test function


- implement read function and tests (single [x], multi-attribute [ ])
- implement custom indexer
- implement oindex...
"""

import random

import numpy as np
from numpy.testing import assert_array_equal

import tiledb
from tiledb.multirange_indexing import getitem_ranges, mr_dense_result_shape
from tiledb.tests.common import DiskTestCase, assert_tail_equal, intspace


def make_1d_dense(path, attr_name="", attr_dtype=np.int64):
    a_orig = np.arange(36)

    dom = tiledb.Domain(tiledb.Dim(domain=(0, 35), tile=35, dtype=np.uint64))
    att = tiledb.Attr(name=attr_name, dtype=attr_dtype)
    schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=False)
    tiledb.DenseArray.create(path, schema)

    with tiledb.DenseArray(path, "w") as A:
        A[:] = a_orig


def make_2d_dense(path, attr_name="", attr_dtype=np.int64):
    a_orig = np.arange(1, 37).reshape(9, 4)

    dom = tiledb.Domain(
        tiledb.Dim(domain=(0, 8), tile=9, dtype=np.uint64),
        tiledb.Dim(domain=(0, 3), tile=4, dtype=np.uint64),
    )
    att = tiledb.Attr(name=attr_name, dtype=attr_dtype)
    schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=False)
    tiledb.DenseArray.create(path, schema)

    with tiledb.DenseArray(path, "w") as A:
        A[:] = a_orig


class TestMultiRangeAuxiliary(DiskTestCase):
    def test_shape_funcs(self):
        range1el = (((1, 1),),)
        self.assertEqual(mr_dense_result_shape(range1el), (1,))

        range1d = tuple([((1, 2), (4, 4))])
        self.assertEqual(mr_dense_result_shape(range1d), (3,))

        range2d1 = (((3, 6), (7, 7), (10, 12)), ((5, 7),))
        self.assertEqual(mr_dense_result_shape(range2d1), (8, 3))

        # range2d2 = ([(3, 6), (7, 7), (10, 12)], [(5, 7), (10, 10)])

    # def test_3d(self):
    #     range3d1 = (((2, 4),), ((3, 6),), ((1, 4), (5, 9)))
    #
    #     # self.assertEqual()

    def test_sel_to_ranges(self):
        class Obj(object):
            pass

        class IBI(object):
            def __getitem__(self, idx):
                return idx

        def make_arr(ndim):
            arr = Obj()
            arr.schema = Obj()
            arr.schema.domain = Obj()
            arr.schema.domain.ndim = ndim
            arr.array = Obj()
            # place-holder for attribute that is not used in these tests
            arr.nonempty_domain = lambda: [()] * ndim
            return arr

        ibi = IBI()
        # ndim = 1
        arr = make_arr(1)
        self.assertEqual(getitem_ranges(arr, ibi[[1]]), (((1, 1),),))
        self.assertEqual(getitem_ranges(arr, ibi[[1, 2]]), (((1, 1), (2, 2)),))
        self.assertEqual(getitem_ranges(arr, ibi[slice(1, 2)]), (((1, 2),),))
        self.assertEqual(getitem_ranges(arr, ibi[1:2]), (((1, 2),),))

        # ndim = 2
        arr2 = make_arr(2)
        self.assertEqual(getitem_ranges(arr2, ibi[[1]]), (((1, 1),), ()))
        self.assertEqual(getitem_ranges(arr2, ibi[slice(1, 33)]), (((1, 33),), ()))
        self.assertEqual(
            getitem_ranges(arr2, ibi[[1, 2], [[1], slice(1, 3)]]),
            (((1, 1), (2, 2)), ((1, 1), (1, 3))),
        )

        # ndim = 3
        arr3 = make_arr(3)
        self.assertEqual(
            getitem_ranges(arr3, ibi[1, 2, 3]), (((1, 1),), ((2, 2),), ((3, 3),))
        )
        self.assertEqual(getitem_ranges(arr3, ibi[1, 2]), ((((1, 1),), ((2, 2),), ())))
        self.assertEqual(
            getitem_ranges(arr3, ibi[1:2, 3:4]), (((1, 2),), ((3, 4),), ())
        )
        self.assertEqual(
            getitem_ranges(arr3, ibi[1:2, 3:4, 5:6]), (((1, 2),), ((3, 4),), ((5, 6),))
        )
        self.assertEqual(
            getitem_ranges(arr3, ibi[[1], [2], [5, 6]]),
            (((1, 1),), ((2, 2),), ((5, 5), (6, 6))),
        )
        self.assertEqual(
            getitem_ranges(arr3, ibi[1, [slice(3, 6), 8], slice(4, 6)]),
            (((1, 1),), ((3, 6), (8, 8)), ((4, 6),)),
        )
        self.assertEqual(getitem_ranges(arr3, ibi[(1, 2)]), (((1, 1),), ((2, 2),), ()))
        self.assertEqual(getitem_ranges(arr3, ibi[[(1, 2)]]), (((1, 2),), (), ()))
        self.assertEqual(
            getitem_ranges(arr3, ibi[[(1, 2), 4], [slice(1, 4)]]),
            (((1, 2), (4, 4)), ((1, 4),), ()),
        )


class TestMultiRange(DiskTestCase):
    def test_multirange_behavior(self):
        uri = self.path("multirange_behavior_sparse")

        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(
                *[
                    tiledb.Dim(
                        name="idx",
                        domain=(-1.0, 0.7999999999999996),
                        tile=2.0,
                        dtype="float64",
                    )
                ]
            ),
            attrs=[tiledb.Attr(name="data", dtype="float64", var=False)],
            cell_order="row-major",
            tile_order="row-major",
            capacity=10000,
            sparse=True,
            allows_duplicates=True,
        )
        tiledb.SparseArray.create(uri, schema)
        data = np.random.rand(10)
        idx = np.arange(-1, 1, 0.2)

        with tiledb.open(uri, "w") as A:
            A[idx] = {"data": data}

        with tiledb.open(uri) as A:
            res = A.multi_index[:]
            # always return data
            self.assertTrue("data" in res)
            # return coordinates for sparse
            self.assertTrue("idx" in res)
            assert_array_equal(res["data"], data)
            assert_array_equal(res["idx"], idx)

        uri = self.path("multirange_behavior_dense")
        with tiledb.from_numpy(uri, data):
            pass

        with tiledb.open(uri) as B:
            res = B.multi_index[0:9]  # TODO: this should accept [:]
            # always return data
            self.assertTrue("" in res)
            # don't return coordinates for dense
            self.assertTrue("idx" not in res)

    def test_multirange_empty(self):
        path1 = self.path("test_multirange_empty_1d")
        make_1d_dense(path1, attr_dtype=np.uint16)
        with tiledb.open(path1) as A:
            res = A.multi_index[tiledb.EmptyRange]
            assert res[""].dtype == np.uint16
            assert res[""].shape == (0,)

        path2 = self.path("test_multirange_empty_2d")
        make_2d_dense(path2, attr_dtype=np.float32)
        with tiledb.open(path2) as A:
            res = A.multi_index[tiledb.EmptyRange]
            assert res[""].dtype == np.float32
            assert res[""].shape == (0,)

    def test_multirange_1d_1dim_ranges(self):
        path = self.path("test_multirange_1d_1dim_ranges")
        attr_name = "a"

        make_1d_dense(path, attr_name=attr_name)

        with tiledb.DenseArray(path) as A:
            ranges = (((0, 0),),)
            expected = np.array([0], dtype=np.int64)
            res = tiledb.libtiledb.multi_index(A, (attr_name,), ranges)
            a = res[attr_name]
            assert_array_equal(a, expected)
            self.assertEqual(a.dtype, expected.dtype)
            self.assertEqual(len(res.keys()), 2)

            ranges2 = (((1, 1), (5, 8)),)
            expected2 = np.array([1, 5, 6, 7, 8], dtype=np.int64)
            a2 = tiledb.libtiledb.multi_index(A, (attr_name,), ranges2)[attr_name]
            assert_array_equal(a2, expected2)
            self.assertEqual(a2.dtype, expected2.dtype)

    def test_multirange_2d_1dim_ranges(self):
        path = self.path("test_multirange_1dim_ranges")
        attr_name = "a"

        make_2d_dense(path, attr_name=attr_name)

        expected = np.array(
            [
                1,
                2,
                3,
                4,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
                28,
                29,
                30,
                31,
                32,
                33,
                34,
                35,
                36,
            ],
            dtype=np.uint64,
        )

        ranges = (((0, 0), (5, 8)),)

        with tiledb.DenseArray(path) as A:
            a = tiledb.libtiledb.multi_index(A, (attr_name,), ranges)[attr_name]

            assert_array_equal(a, expected)

    def test_multirange_2d_2dim_ranges(self):
        path = self.path("test_multirange_2dim_ranges")
        attr_name = "a"

        make_2d_dense(path, attr_name=attr_name)

        expected = np.arange(1, 21)

        ranges = (((0, 4),), ((0, 3),))

        with tiledb.DenseArray(path) as A:
            a = tiledb.libtiledb.multi_index(A, (attr_name,), ranges)[attr_name]
            assert_array_equal(a, expected)

            # test slicing start=end on 1st dim at 0 (bug fix)
            assert_tail_equal(
                np.array([[1, 2, 3, 4]]),
                A.multi_index[:0][attr_name],
                A.multi_index[0:0][attr_name],
            )

            # test slicing start=end on 2nd dim at 0 (bug fix)
            assert_tail_equal(
                np.arange(1, 34, 4).reshape((9, 1)),
                A.multi_index[:, :0][attr_name],
                A.multi_index[:, 0:0][attr_name],
            )

            # test slicing start=end on 1st dim at 1
            assert_array_equal(np.array([[5, 6, 7, 8]]), A.multi_index[1:1][attr_name])

            # test slicing start=end on 2nd dim at 1
            assert_array_equal(
                np.arange(2, 35, 4).reshape((9, 1)), A.multi_index[:, 1:1][attr_name]
            )

            # test slicing start=end on 1st dim at max range
            assert_array_equal(
                np.array([[33, 34, 35, 36]]), A.multi_index[8:8][attr_name]
            )

            # test slicing start=end on 2nd dim at max range
            assert_tail_equal(
                np.arange(4, 37, 4).reshape((9, 1)), A.multi_index[:, 3:3][attr_name]
            )

    def test_multirange_1d_dense_int64(self):
        attr_name = ""
        path = self.path("multi_index_1d")

        dom = tiledb.Domain(
            tiledb.Dim(name="coords", domain=(-10, 10), tile=9, dtype=np.int64),
        )
        att = tiledb.Attr(name=attr_name, dtype=np.float32)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,))
        tiledb.DenseArray.create(path, schema)

        orig_array = np.random.rand(schema.domain.dim(0).size).astype(np.float32)
        with tiledb.open(path, "w") as A:
            A[:] = orig_array

        with tiledb.open(path) as A:
            # stepped ranges are not supported
            with self.assertRaises(ValueError):
                A.query(coords=True).multi_index[1::2]

            assert_array_equal(orig_array[[0, -1]], A.multi_index[[-10, 10]][attr_name])
            self.assertEqual(orig_array[0], A.multi_index[-10][attr_name])
            self.assertEqual(
                -10, A.query(coords=True).multi_index[-10]["coords"].view("i8")
            )
            assert_array_equal(orig_array[0:], A.multi_index[[(-10, 10)]][attr_name])
            assert_array_equal(
                orig_array[0:], A.multi_index[[slice(-10, 10)]][attr_name]
            )
            assert_array_equal(
                orig_array[0:10], A.multi_index[-10 : np.int64(-1)][attr_name]
            )
            assert_array_equal(orig_array, A.multi_index[:][attr_name])
            ned = A.nonempty_domain()[0]
            assert_array_equal(
                A.multi_index[ned[0] : ned[1]][attr_name], A.multi_index[:][attr_name]
            )

    def test_multirange_1d_sparse_double(self):
        attr_name = ""
        path = self.path("mr_1d_sparse_double")

        dom = tiledb.Domain(
            tiledb.Dim(name="coords", domain=(0, 30), tile=10, dtype=np.float64),
        )
        att = tiledb.Attr(name=attr_name, dtype=np.float64)
        schema = tiledb.ArraySchema(domain=dom, sparse=True, attrs=(att,))
        tiledb.SparseArray.create(path, schema)

        coords = np.linspace(0, 30, num=31)
        orig_array = np.random.rand(coords.size)

        with tiledb.open(path, "w") as A:
            A[coords] = orig_array

        with tiledb.open(path) as A:
            assert_array_equal(orig_array[[0]], A.multi_index[[0]][attr_name])
            assert_array_equal(orig_array[-1], A.multi_index[30][attr_name])
            assert_array_equal(orig_array[-1], A.multi_index[30.0][attr_name])
            assert_array_equal(
                orig_array[coords.size - 3 : coords.size],
                A.multi_index[
                    (28.0, 30.0),
                ][attr_name],
            )

            res = A.multi_index[slice(0, 5)]
            assert_array_equal(orig_array[0:6], res[attr_name])
            assert_array_equal(coords[0:6], res["coords"].astype(np.float64))

            # test slice range indexing
            ned = A.nonempty_domain()
            res = A.multi_index[: ned[0][1]]
            assert_array_equal(coords, res["coords"].astype(np.float64))

            res = A.multi_index[ned[0][0] : coords[15]]
            assert_array_equal(coords[:16], res["coords"].astype(np.float64))

    def test_multirange_2d_sparse_domain_utypes(self):
        attr_name = "foo"

        types = (np.uint8, np.uint16, np.uint32, np.uint64)

        for dtype in types:
            min = 0
            max = int(np.iinfo(dtype).max) - 1
            path = self.path("multi_index_2d_sparse_" + str(dtype.__name__))

            dom = tiledb.Domain(tiledb.Dim(domain=(min, max), tile=1, dtype=dtype))

            att = tiledb.Attr(name=attr_name, dtype=dtype)
            schema = tiledb.ArraySchema(domain=dom, sparse=True, attrs=(att,))
            tiledb.SparseArray.create(path, schema)

            coords = intspace(min, max, num=100, dtype=dtype)

            with tiledb.open(path, "w") as A:
                A[coords] = coords

            with tiledb.open(path) as A:

                res = A.multi_index[slice(coords[0], coords[-1])]
                assert_array_equal(res[attr_name], coords)
                assert_array_equal(res["__dim_0"].astype(dtype), coords)

                res = A.multi_index[coords[0]]
                assert_array_equal(res[attr_name], coords[0])
                assert_array_equal(res["__dim_0"].astype(dtype), coords[0])

                res = A.multi_index[coords[-1]]
                assert_array_equal(res[attr_name], coords[-1])
                assert_array_equal(res["__dim_0"].astype(dtype), coords[-1])

                midpoint = len(coords) // 2
                start = midpoint - 20
                stop = midpoint + 20
                srange = slice(coords[start], coords[stop])
                res = A.multi_index[srange]
                assert_array_equal(res[attr_name], coords[start : stop + 1])
                assert_array_equal(
                    res["__dim_0"].astype(dtype), coords[start : stop + 1]
                )

    def test_multirange_2d_sparse_float(self):
        attr_name = ""
        path = self.path("mr_2d_sparse_float")

        dom = tiledb.Domain(
            tiledb.Dim(domain=(0, 10), tile=1, dtype=np.float32),
            tiledb.Dim(domain=(0, 10), tile=1, dtype=np.float32),
        )
        att = tiledb.Attr(name=attr_name, dtype=np.float64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=True)
        tiledb.SparseArray.create(path, schema)

        orig_array = np.random.rand(11, 11)
        d1 = np.linspace(0, 10, num=11, dtype=np.float32)
        d2 = np.linspace(0, 10, num=11, dtype=np.float32)
        coords_d1, coords_d2 = np.meshgrid(d1, d2, indexing="ij")

        with tiledb.open(path, "w") as A:
            A[coords_d1.flatten(), coords_d2.flatten()] = orig_array

        with tiledb.open(path) as A:
            res = A.multi_index[[0], :]
            assert_array_equal(orig_array[[0], :].squeeze(), res[attr_name])
            assert_array_equal(coords_d1[0, :], res["__dim_0"])

            # ===
            res = A.multi_index[10, :]
            assert_array_equal(orig_array[[-1], :].squeeze(), res[attr_name])
            assert_array_equal(coords_d2[[-1], :].squeeze(), res["__dim_1"])

            # ===
            res = A.multi_index[[slice(0, 2), [5]]]
            assert_array_equal(
                np.vstack([orig_array[0:3, :], orig_array[5, :]]).flatten(),
                res[attr_name],
            )
            assert_array_equal(
                np.vstack((coords_d1[0:3], coords_d1[5])).flatten(), res["__dim_0"]
            )

            # ===
            res = A.multi_index[slice(0.0, 2.0), slice(2.0, 5.0)]
            assert_array_equal(orig_array[0:3, 2:6].flatten(), res[attr_name])
            assert_array_equal(coords_d1[0:3, 2:6].flatten(), res["__dim_0"])
            assert_array_equal(coords_d2[0:3, 2:6].flatten(), res["__dim_1"])
            res = A.multi_index[
                slice(np.float32(0.0), np.float32(2.0)),
                slice(np.float32(2.0), np.float32(5.0)),
            ]
            assert_array_equal(orig_array[0:3, 2:6].flatten(), res[attr_name])
            assert_array_equal(coords_d1[0:3, 2:6].flatten(), res["__dim_0"])
            assert_array_equal(coords_d2[0:3, 2:6].flatten(), res["__dim_1"])

    def test_multirange_1d_sparse_query(self):
        path = self.path("mr_1d_sparse_query")

        dom = tiledb.Domain(
            tiledb.Dim(name="coords", domain=(-100, 100), tile=1, dtype=np.float32),
        )
        attrs = [
            tiledb.Attr(name="U", dtype=np.float64),
            tiledb.Attr(name="V", dtype=np.uint32),
        ]

        schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=True)
        tiledb.SparseArray.create(path, schema)

        U = np.random.rand(11)
        V = np.random.randint(0, np.iinfo(np.uint32).max, 11, dtype=np.uint32)

        coords = np.linspace(-10, 10, num=11, dtype=np.float32)
        data = {"U": U, "V": V}

        with tiledb.open(path, "w") as A:
            A[coords] = data

        with tiledb.open(path) as A:
            for k, d in data.items():
                Q = A.query(attrs=k)

                res = Q.multi_index[[-10]]
                assert_array_equal(d[[0]], res[k])

                assert_array_equal(coords[[0]], res["coords"].view("f4"))

                res = A.multi_index[10]
                assert_array_equal(d[[-1]].squeeze(), res[k])

                assert_array_equal(coords[[-1]], res["coords"].view("f4"))

                res = A.multi_index[[slice(coords[0], coords[2]), [coords[-1]]]]
                assert_array_equal(np.hstack([d[0:3], d[-1]]), res[k])

                # make sure full slice indexing works on query
                res = Q.multi_index[:]
                assert_array_equal(coords, res["coords"])

                # TODO: this should be an error
                # res = A.multi_index[10, :]
                # assert_array_equal(
                #    d[[-1]].squeeze(),
                #    res[k]
                # )

        with tiledb.open(path) as A:
            Q = A.query(coords=False, attrs=["U"])
            res = Q.multi_index[:]
            self.assertTrue("U" in res)
            self.assertTrue("V" not in res)
            self.assertTrue("coords" not in res)
            assert_array_equal(res["U"], data["U"])

    def test_multirange_1d_dense_vectorized(self):
        path = self.path("mr_1d_dense_vectorized")

        dom = tiledb.Domain(tiledb.Dim(domain=(0, 999), tile=1000, dtype=np.uint32))
        attrs = tiledb.Attr(name="", dtype=np.float64)

        schema = tiledb.ArraySchema(domain=dom, attrs=(attrs,), sparse=False)
        tiledb.DenseArray.create(path, schema)

        data = np.random.rand(1000)
        with tiledb.DenseArray(path, "w") as A:
            A[0] = data[0]
            A[-1] = data[-1]
            A[:] = data

        for _ in range(0, 50):
            with tiledb.DenseArray(path) as A:
                idxs = random.sample(range(0, 999), k=100)
                res = A.multi_index[idxs]
                assert_array_equal(data[idxs], res[""])

    def test_multirange_2d_dense_float(self):
        attr_name = ""
        path = self.path("multirange_2d_dense_float")

        dom = tiledb.Domain(
            tiledb.Dim(domain=(0, 10), tile=1, dtype=np.int64),
            tiledb.Dim(domain=(0, 10), tile=1, dtype=np.int64),
        )
        att = tiledb.Attr(name=attr_name, dtype=np.float64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=False)
        tiledb.DenseArray.create(path, schema)

        orig_array = np.random.rand(11, 11)

        with tiledb.open(path, "w") as A:
            A[:] = orig_array

        with tiledb.open(path) as A:
            assert_array_equal(orig_array[[0], :], A.multi_index[[0], :][attr_name])
            assert_array_equal(
                orig_array[[-1, -1], :], A.multi_index[[10, 10], :][attr_name]
            )
            assert_array_equal(
                orig_array[0:4, 7:10], A.multi_index[[(0, 3)], slice(7, 9)][attr_name]
            )
            assert_array_equal(orig_array[:, :], A.multi_index[:, :][attr_name])
            # TODO this should be an error to match NumPy 1.12 semantics
            # assert_array_equal(
            #    orig_array[0:4,7:10],
            #    A.multi_index[[(np.float64(0),np.float64(3.0))], slice(7,9)][attr_name]
            # )

    def test_multirange_1d_sparse_datetime64(self):
        path = self.path("multirange_1d_sparse_datetime64")

        dom = tiledb.Domain(
            tiledb.Dim(
                domain=(np.datetime64("2019"), np.datetime64("2020")),
                dtype="datetime64[D]",
                tile=1,
            ),
        )

        attr_name = ""
        att = tiledb.Attr(name=attr_name, dtype=np.float64)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=True)
        tiledb.SparseArray.create(path, schema)

        with tiledb.SparseArray(path, mode="w") as T:
            dates = np.array(
                ["2019-10-02", "2019-10-03", "2019-10-04"], dtype="datetime64[D]"
            )
            T[dates] = np.array(range(3))

        with tiledb.open(path) as A:
            oct_2_np = np.datetime64("2019-10-02").astype("int64")
            oct_4_np = np.datetime64("2019-10-04").astype("int64")

            oct_2_int = oct_2_np.item()
            oct_4_int = oct_4_np.item()

            self.assertEqual(oct_2_np, oct_2_int)
            self.assertEqual(oct_4_np, oct_4_int)

            assert_array_equal(
                (A.multi_index[oct_2_np:oct_4_np][attr_name]),
                A.multi_index[oct_2_int:oct_4_int][attr_name],
            )

    def test_fix_473_sparse_index_bug(self):
        # test of fix for issue raised in
        # https://github.com/TileDB-Inc/TileDB-Py/pull/473#issuecomment-784675012

        uri = self.path("test_fix_473_sparse_index_bug")
        dom = tiledb.Domain(
            tiledb.Dim(name="x", domain=(0, 2 ** 64 - 2), tile=1, dtype=np.uint64)
        )
        schema = tiledb.ArraySchema(
            domain=dom, sparse=True, attrs=[tiledb.Attr(name="a", dtype=np.uint64)]
        )

        tiledb.SparseArray.create(uri, schema)

        slice_index = slice(0, 4, None)

        with tiledb.SparseArray(uri, mode="r") as A:
            data = A.multi_index[slice_index]

            assert_array_equal(data["a"], np.array([], dtype=np.uint64))
            assert_array_equal(A.multi_index[:], [])

        with tiledb.open(uri, mode="w") as A:
            A[[10]] = {"a": [10]}

        with tiledb.open(uri) as A:
            assert_tail_equal(
                A.multi_index[slice_index]["a"],
                A.multi_index[:],
                A.multi_index[0:],
                A.multi_index[1:],
                A.multi_index[:10],
                A.multi_index[:11],
                np.array([], dtype=np.uint64),
            )

    def test_fixed_multi_attr_df(self):
        uri = self.path("test_fixed_multi_attr_df")
        dom = tiledb.Domain(
            tiledb.Dim(name="dim", domain=(0, 0), tile=None, dtype=np.int32),
        )
        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=True,
            attrs=[
                tiledb.Attr(
                    name="111",
                    dtype=[("", np.int32), ("", np.int32), ("", np.int32)],
                )
            ],
        )

        tiledb.SparseArray.create(uri, schema)

        data_111 = np.array(
            [(1, 1, 1)], dtype=[("", np.int32), ("", np.int32), ("", np.int32)]
        )
        with tiledb.SparseArray(uri, mode="w") as A:
            A[0] = data_111

        with tiledb.SparseArray(uri, mode="r") as A:
            result = A.query(attrs=["111"])[0]
            assert_array_equal(result["111"], data_111)

            with self.assertRaises(tiledb.TileDBError):
                result = A.query(attrs=["111"]).df[0]

            result = A.query(attrs=["111"], use_arrow=False)
            assert_array_equal(result.df[0]["111"], data_111)

    def test_var_multi_attr_df(self):
        uri = self.path("test_var_multi_attr_df")
        dom = tiledb.Domain(
            tiledb.Dim(name="dim", domain=(0, 2), tile=None, dtype=np.int32),
        )
        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=True,
            attrs=[tiledb.Attr(name="1s", dtype=np.int32, var=True)],
        )

        tiledb.SparseArray.create(uri, schema)

        data = np.array(
            [
                np.array([1], dtype=np.int32),
                np.array([1, 1], dtype=np.int32),
                np.array([1, 1, 1], dtype=np.int32),
            ],
            dtype="O",
        )
        with tiledb.SparseArray(uri, mode="w") as A:
            A[[0, 1, 2]] = data

        with tiledb.SparseArray(uri, mode="r") as A:
            result = A.query(attrs=["1s"])
            assert_array_equal(result[0]["1s"][0], data[0])
            assert_array_equal(result[1]["1s"][0], data[1])
            assert_array_equal(result[2]["1s"][0], data[2])

            with self.assertRaises(tiledb.TileDBError):
                result = A.query(attrs=["1s"]).df[0]

            result = A.query(attrs=["1s"], use_arrow=False)
            assert_array_equal(result.df[0]["1s"][0], data[0])
            assert_array_equal(result.df[1]["1s"][0], data[1])
            assert_array_equal(result.df[2]["1s"][0], data[2])

    def test_multi_index_with_implicit_full_string_range(self):
        uri = self.path("test_multi_index_with_implicit_full_string_range")
        dom = tiledb.Domain(
            tiledb.Dim(name="dint", domain=(0, 4), tile=5, dtype=np.int32),
            tiledb.Dim(name="dstr", domain=(None, None), tile=None, dtype=np.bytes_),
        )
        schema = tiledb.ArraySchema(
            domain=dom,
            sparse=True,
            attrs=[tiledb.Attr(name="", dtype=np.int32)],
        )

        tiledb.Array.create(uri, schema)
        with tiledb.open(uri, mode="w") as A:
            d1 = np.concatenate((np.arange(5), np.arange(5)))
            d2 = np.asarray(
                ["a", "b", "ab", "ab", "c", "c", "c", "c", "d", "e"], dtype=np.bytes_
            )
            A[d1, d2] = np.array(np.random.randint(10, size=10), dtype=np.int32)

        with tiledb.open(uri, mode="r") as A:
            assert_array_equal(A[:][""], A.multi_index[:][""])
            assert_array_equal(A.multi_index[:][""], A.multi_index[:, :][""])

            assert_array_equal(A[1:4][""], A.multi_index[1:3][""])
            assert_array_equal(A.multi_index[1:3][""], A.multi_index[1:3, :][""])

            assert_array_equal(A[0][""], A.multi_index[0][""])
            assert_array_equal(A.multi_index[0][""], A.multi_index[0, :][""])

    def test_multi_index_open_timestamp_with_empty_nonempty_domain(self):
        uri = self.path("test_multi_index_open_timestamp_with_empty_nonempty_domain")
        dom = tiledb.Domain(tiledb.Dim(domain=(1, 3)))
        attr = tiledb.Attr(name="", dtype=np.int32)
        schema = tiledb.ArraySchema(domain=dom, sparse=True, attrs=[attr])
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, mode="w", timestamp=2) as A:
            d1 = np.array(np.random.randint(1, 11, size=3, dtype=np.int32))
            A[np.arange(1, 4)] = d1

        with tiledb.open(uri, mode="r", timestamp=1) as A:
            assert A.nonempty_domain() is None
            assert_array_equal(A.multi_index[:][""], A[:][""])



import tiledb
import numpy as np
import pandas as pd
import pandas._testing as tm

import hypothesis
import hypothesis.strategies as st
from hypothesis import given
from numpy.testing import assert_array_equal

from tiledb.tests.common import DiskTestCase


class AttrDataTest(DiskTestCase):
    @hypothesis.settings(deadline=1000)
    @given(st.binary())
    def test_bytes_df(self, data):
        # TODO this test is slow. might be nice to run with in-memory
        #      VFS (if faster) but need to figure out correct setup
        # uri = "mem://" + str(uri_int)

        uri = self.path()
        uri_df = self.path()

        if data == b"" or data.count(b"\x00") == len(data):
            # single-cell empty writes are not supported; TileDB PR 1646
            array = np.array([data, b"1"], dtype="S0")
        else:
            array = np.array([data], dtype="S0")

        # DEBUG
        tiledb.stats_enable()
        tiledb.stats_reset()
        # END DEBUG

        with tiledb.from_numpy(uri, array) as A:
            pass

        with tiledb.open(uri) as A:
            assert_array_equal(A.multi_index[:][""], array)

        hypothesis.note(tiledb.stats_dump(print_out=False))

        # DEBUG
        tiledb.stats_disable()

    @hypothesis.settings(deadline=1000)
    @given(st.binary())
    def test_bytes_numpy(self, data):
        # TODO this test is slow. might be nice to run with in-memory
        #      VFS (if faster) but need to figure out correct setup
        # uri = "mem://" + str(uri_int)

        uri = self.path()
        uri_df = self.path()

        if data == b"" or data.count(b"\x00") == len(data):
            # single-cell empty writes are not supported; TileDB PR 1646
            array = np.array([data, b"1"], dtype="S0")
        else:
            array = np.array([data], dtype="S0")

        series = pd.Series(array)
        df = pd.DataFrame({"": series})

        # DEBUG
        tiledb.stats_enable()
        tiledb.stats_reset()
        # END DEBUG

        tiledb.from_pandas(uri_df, df, sparse=False)

        with tiledb.open(uri_df) as A:
            tm.assert_frame_equal(A.df[:], df)

        hypothesis.note(tiledb.stats_dump(print_out=False))

        # DEBUG
        tiledb.stats_disable()
